%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% (CHEM-E7225) ADVANCED PROCESS CONTROL D
%   #03 - Optimal control of dynamic systems
%
%   Complete this starting code (and functions ThreeTank.m and rk4fnc.m) to control 
%   the Three-Tank System for the control tasks described in Exercise 03.
%
%   Find and replace each '???' with the correct code. 
%
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
close all; clear; clc

addpath('~/path/to/casadi')
import casadi.*

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% DYNAMICAL SYSTEM DEFINITIONS
% Control horizon and discretization
K  = ???;          % Control horizon
dT = ???;          % Discretisation period
t  = ???;          % Time nodes (t_0,...,t_K-1)

% Dynamics, initial and reference conditions
Nx = ???;          % Number of state vars
Nu = ???;          % Number of controls

% Model parameters
S_T = ???;  S_V = ???;  % and so on...

px = [ ??? ];

% Initial state
x0 = [ ??? ];

% NONLINEAR MODEL: Integrator, using RK4
%  Differently from other Homeworks, here we use CasADi
%  to create the integrator function. This has better
%  performance than defining F = @(x,u) rk4fnc(...)
X = MX.sym('x',Nx);
U = MX.sym('u',Nu);

F = Function('F', {X,U}, {rk4fnc(@ThreeTank, dT, X, U, px)});

%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% OPTIMAL CONTROL PROBLEM DEFINITIONS

% == OPTIMIZATION DATA ==
% Construct state-reference functions
x1_ref = ???;   % <~ Fill with the correct expression for x_1^ref (k=0,...,K-1)
x2_ref = ???;   % <~ Fill with the correct expression for x_2^ref (k=0,...,K-1)
x3_ref = ???;   % <~ Fill with the correct expression for x_3^ref (k=0,...,K-1)

x_ref = [x1_ref; x2_ref; x3_ref];

% Optimisation parameters
Q = ???;        % <~ Fill with the state-tracking weights
R = ???;        % <~ Fill with the input weights

% Define lower and upper bounds
X_lb = ???;     % <~ Fill with the states lower bounds (for all k=0,...,K)
X_ub = ???;     % <~ Fill with the states upper bounds (for all k=0,...,K)

U_lb = ???;     % <~ Fill with the controls lower bounds (for all k=0,...,K-1)
U_ub = ???;     % <~ Fill with the controls upper bounds (for all k=0,...,K-1)

% == OPTIMIZATION PROBLEM ==
opti = casadi.Opti();             % Creates the optimization problem environment

X = opti.variable( ??? );         % <~ (Decision vars) Fill the states dimensions
U = opti.variable( ??? );         % <~ (Decision vars) Fill the controls dimensions

% - Simulates the system, computes cost function, and store constraints
sumLx = 0; sumLu = 0;  g_x = [];    % Objective function (J = sumLx + sumLu) and constraints (g_x)
for k = 1:K-1
    sumLx = sumLx + ???;            % Compute state cost, at stage k
    sumLu = sumLu + ???;            % Compute control cost, at stage k
    
    g_x = [g_x; ???];               % Add dynamic equality constraint x_(k+1) - F(x_k,u_k) = 0
end

J = (sumLx + sumLu);              % Computes the objective function (without terminal cost)
% --

% Updates the Opti Stack environment with the objective and constraints
opti.minimize( ??? )                        % <~ Fill with the objective function to be minimized, J(x,u)
opti.subject_to( ??? )                      % <~ Fill with the dynamic equality constraints, g(x) = 0
opti.subject_to( ??? )                      % <~ Fill with the state bounds, x^min <= x <= x^max
opti.subject_to( ??? )                      % <~ Fill with the control bounds, u^min <= u <= u^max
opti.subject_to( ??? )                      % <~ Fill with the Initial state constraint, x_0 = \hat{x}_0

opti.solver('ipopt')                        % Solver (IPOPT, qpOASES, ...)
opti.set_initial(X, rand(Nx,K))             % (States) Random initial solution
opti.set_initial(U, rand(Nu,K-1))           % (Controls) random initial solution
opti.solve();                               % Executes the solver

% == ==

x_sol = opti.value(X);                      % Retrieves the optimal solution of (x)
u_sol = opti.value(U);                      % Retrieves the optimal solution of (u)

% Saves the solution (to be plotted in E03B_main.m)
save('ThreeTank_OCP.mat', 'x_sol', 'u_sol')   


%% %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% Plotting all the simulation results
%
%  NOTE: You don't need to complete this part, but feel free to 
%        customize the plots according to your style. 

lw = 1.0; fs = 14;
figure(1), clf; 
for j = 1:2
  subplot(2,3,j+(j==2)), 
    stairs(t(1:end-1), u_sol(j,:), 'k-', 'linewidth', lw)
    ylabel("$u_"+j+"(t)$ [ml/s]",'Interpreter','Latex'); ylim([0,140])
    set(gca,'TickLabelInterpreter','Latex','FontSize',fs)
    box off
end

for j = 1:3
  subplot(2,3,3+j), hold on
    stairs(t,x_ref(j,:)', 'r-', 'linewidth', lw); 
    stairs(t,x_sol(j,:)', 'k-', 'linewidth', lw); 
    ylabel("$x_"+j+"(t)$ [cm]",'Interpreter','Latex'); ylim([0,60])
    xlabel('Time  - $t$ [s]','Interpreter','Latex')
    set(gca,'TickLabelInterpreter','Latex','FontSize',fs)
    box off

    if j == 3; legend(["Reference"], 'Interpreter', 'Latex', 'Box', 'Off')
    end
end
